# -*- coding: utf-8 -*-
import time
from pprint import pformat
from gi.repository import GLib, Gio, Gtk

from dfeet.uiloader import UILoader


class ExecuteMethodDialog:
    def __init__(self, data_dir, connection, connection_is_bus, bus_name,
                 method_obj, parent_window):
        signal_dict = {
            'execute_dbus_method_cb': self.execute_cb,
            'execute_dialog_close_cb': self.close_cb
            }

        self.connection = connection
        self.connection_is_bus = connection_is_bus
        self.bus_name = bus_name
        self.method_obj = method_obj

        ui = UILoader(data_dir, UILoader.UI_EXECUTEDIALOG)
        self.dialog = ui.get_root_widget()
        self.dialog.set_transient_for(parent_window)
        self.label_method_name = ui.get_widget('label_method_name')
        self.label_object_path = ui.get_widget('label_object_path')
        self.label_interface = ui.get_widget('label_interface')
        self.notebook = ui.get_widget('notebook1')
        self.parameter_textview = ui.get_widget('parametertextview1')
        self.source_textview = ui.get_widget('sourcetextview1')
        self.prettyprint_textview = ui.get_widget('prettyprinttextview1')
        self.method_execution_count_spinbutton = ui.get_widget('method_exec_count_spinbutton')
        self.label_avg = ui.get_widget('label_avg')
        self.label_min = ui.get_widget('label_min')
        self.label_max = ui.get_widget('label_max')
        ui.connect_signals(signal_dict)
        self.label_method_name.set_markup("%s" % (self.method_obj.markup_str))
        self.label_object_path.set_markup("%s" % (self.method_obj.object_path))
        self.label_interface.set_markup("%s" % (self.method_obj.iface_info.name))

    def execute_cb(self, widget):
        # get given parameters
        buf = self.parameter_textview.get_buffer()
        params = buf.get_text(buf.get_start_iter(),
                              buf.get_end_iter(), False)

        # reset the statistics stuff
        self.label_avg.set_text("")
        self.label_min.set_text("")
        self.label_max.set_text("")
        user_data = {
            'avg': 0,
            'count': 0,
            }

        try:
            # build a GVariant
            if params:
                params = "(" + params + ",)"
                params_code = '(' + self.method_obj.in_args_code + ')'
                params_gvariant = GLib.Variant(params_code, eval(params))
            else:
                params_gvariant = None

            if self.connection_is_bus:
                proxy = Gio.DBusProxy.new_sync(self.connection,
                                               Gio.DBusProxyFlags.NONE,
                                               None,
                                               self.bus_name,
                                               self.method_obj.object_path,
                                               self.method_obj.iface_info.name,
                                               None)
                # call the function
                for i in range(0, self.method_execution_count_spinbutton.get_value_as_int()):
                    user_data['method_call_time_start'] = time.time()
                    proxy.call(
                        self.method_obj.method_info.name, params_gvariant,
                        Gio.DBusCallFlags.NONE, -1, None, self.method_connection_bus_cb, user_data)
            else:
                # FIXME: implement p2p connection execution
                raise Exception("Function execution on p2p connections not yet implemented")
                # self.connection.call(
                # None, object_path, self.method_obj.iface_obj.iface_info.name,
                # self.method_obj.method_info.name, params_gvariant,
                # GLib.VariantType.new("(s)"), Gio.DBusCallFlags.NONE, -1, None)

        except Exception as e:
            # output the exception
            self.source_textview.get_buffer().set_text(str(e))
            self.prettyprint_textview.get_buffer().set_text(pformat(str(e)))

    def method_connection_bus_cb(self, proxy, res_async, user_data):
        """async callback for executed method"""
        try:
            # get the result from the dbus method call
            result = proxy.call_finish(res_async)
            # remember the needed time for the method call
            method_call_time_end = time.time()
            method_call_time_needed = method_call_time_end - user_data['method_call_time_start']

            # update avg, min, max
            user_data['avg'] += method_call_time_needed
            user_data['count'] += 1
            self.label_avg.set_text("%.4f" % (float(user_data['avg'] / user_data['count'])))
            self.label_min.set_text(
                "%.4f" % min(float(self.label_min.get_text() or "999"), method_call_time_needed))
            self.label_max.set_text(
                "%.4f" % max(float(self.label_max.get_text() or "0"), method_call_time_needed))

            # output result
            if result:
                self.source_textview.get_buffer().set_text(str(result))
                lines = [pformat(x) for x in result.unpack()]
                self.prettyprint_textview.get_buffer().set_text(',\n'.join(lines))
            else:
                self.prettyprint_textview.get_buffer().set_text(
                    'This method did not return anything')
        except Exception as e:
            # output the exception
            self.source_textview.get_buffer().set_text(str(e))
            self.prettyprint_textview.get_buffer().set_text(pformat(str(e)))

    def run(self):
        response = self.dialog.run()
        if response == Gtk.ResponseType.DELETE_EVENT or response == Gtk.ResponseType.CLOSE:
            self.dialog.destroy()

    def close_cb(self, widget):
        self.dialog.destroy()



# -*- coding: utf-8 -*-
from __future__ import print_function


def convert_complex_type(subsig):
    result = None
    len_consumed = 0

    c = subsig[0]

    c_lookahead = ''
    try:
        c_lookahead = subsig[1]
    except:
        c_lookahead = ''

    if c == 'a' and c_lookahead == '{':  # handle dicts as a special case array
        ss = subsig[2:]
        # account for the trailing '}'
        len_consumed = 3
        c = ss[0]
        key = convert_simple_type(c)

        ss = ss[1:]

        (r, lc) = convert_complex_type(ss)
        if r:
            subtypelist = [key]
            for item in r:
                subtypelist.append(item)

            len_consumed += lc + 1
        else:
            value = convert_simple_type(ss[0])
            subtypelist = [key, value]
            len_consumed += 1

        result = ['Dict of {', subtypelist, '}']

    elif c == 'a':                       # handle an array
        ss = subsig[1:]
        (r, lc) = convert_complex_type(ss)
        if r:
            subtypelist = r
            len_consumed = lc + 1
        else:
            subtypelist = sig_to_type_list(ss[0])
            len_consumed = 1

        result = ['Array of [', subtypelist, ']']
    elif c == '(':                       # handle structs
        # iterate over sig until paren_count == 0
        paren_count = 1
        i = 0
        ss = subsig[1:]
        len_ss = len(ss)
        while i < len_ss and paren_count != 0:
            if ss[i] == '(':
                paren_count += 1
            elif ss[i] == ')':
                paren_count -= 1

            i += 1

        len_consumed = i
        ss = ss[0:i - 1]
        result = ['Struct of (', sig_to_type_list(ss), ')']

    return (result, len_consumed)


def convert_simple_type(c):
    result = None

    if c == 'n':
        result = 'Int16'
    elif c == 'q':
        result = 'UInt16'
    elif c == 'i':
        result = 'Int32'
    elif c == 'u':
        result = 'UInt32'
    elif c == 'x':
        result = 'Int64'
    elif c == 't':
        result = 'UInt64'
    elif c == 's':
        result = 'String'
    elif c == 'b':
        result = 'Boolean'
    elif c == 'y':
        result = 'Byte'
    elif c == 'o':
        result = 'Object Path'
    elif c == 'g':
        result = 'Signature'
    elif c == 'd':
        result = 'Double'
    elif c == 'v':
        result = 'Variant'

    return result


def sig_to_type_list(sig):
    i = 0
    result = []

    sig_len = len(sig)
    while i < sig_len:
        c = sig[i]
        type_ = convert_simple_type(c)
        if not type_:
            (type_, len_consumed) = convert_complex_type(sig[i:])
            if not type_:
                type_ = 'Error(' + c + ')'

            i += len_consumed

        if isinstance(type_, list):
            for item in type_:
                result.append(item)
        else:
                result.append(type_)

        i += 1

    return result


def type_list_to_string(type_list):
    result = ''
    add_cap = False

    for dbus_type in type_list:
        if isinstance(dbus_type, list):
            result += type_list_to_string(dbus_type)
            add_cap = True
        else:
            # we get rid of the leading comma later
            if not add_cap:
                result += ', '
            else:
                add_cap = False

            try:
                result += dbus_type
            except:
                print(type_list)

    return result[2:]


def sig_to_markup(sig, span_attr_str):
    list_ = sig_to_type_list(sig)
    markedup_list = []
    m = '<span ' + span_attr_str + '>'
    m += type_list_to_string(list_)
    m += '</span>'

    return m


def sig_to_string(sig):
    return type_list_to_string(sig_to_type_list(sig))



# -*- coding: utf-8 -*-
from gi.repository import Gtk, Gio
from dfeet.uiloader import UILoader


class AddConnectionDialog:

    def __init__(self, data_dir, parent, address_bus_history=[]):
        ui = UILoader(data_dir, UILoader.UI_ADDCONNECTIONDIALOG)

        self.dialog = ui.get_root_widget()
        self.dialog.set_transient_for(parent)
        # get the hbox and add address combo box with model
        hbox1 = ui.get_widget('hbox1')
        self.address_combo_box_store = Gtk.ListStore(str)
        self.address_combo_box = Gtk.ComboBox.new_with_model_and_entry(
            self.address_combo_box_store)
        self.address_combo_box.set_entry_text_column(0)
        self.label_status = ui.get_widget('label_status')

        hbox1.pack_start(self.address_combo_box, True, True, 0)
        hbox1.show_all()

        # add history to model
        for el in address_bus_history:
            self.address_combo_box_store.append([el])

        self.dialog.add_button('gtk-cancel', Gtk.ResponseType.CANCEL)
        self.dialog.add_button('gtk-connect', Gtk.ResponseType.OK)

    @property
    def address(self):
        tree_iter = self.address_combo_box.get_active_iter()
        if tree_iter is not None:
            model = self.address_combo_box.get_model()
            return model[tree_iter][0]
        else:
            entry = self.address_combo_box.get_child()
            return entry.get_text()

    def run(self):
        response = self.dialog.run()
        if response == Gtk.ResponseType.CANCEL:
            return response
        elif response == Gtk.ResponseType.OK:
            # check if given address is valid
            try:
                is_supported = Gio.dbus_is_supported_address(self.address)
            except Exception as e:
                self.label_status.set_text(str(e))
                self.run()
            else:
                return Gtk.ResponseType.OK

    def destroy(self):
        self.dialog.destroy()



#
#    THIS FILE IS PART OF THE D-FEET PROJECT AND LICENSED UNDER THE GPL. SEE
#    THE 'COPYING' FILE FOR DETAILS
#
#   portions taken from the Jokosher project
#
try:
    # python2.x
    import ConfigParser as configparser
except:
    # python3
    import configparser
import os
import re


class ConfigTokenizer():
    COMMA = re.compile(",")
    FALLTHROUGH = re.compile('(?:[^,.])+')
    # FIXME: String re does not ignore escaped quotes (e.g. \") correctly
    STRING = re.compile('"((?:[^\\"]|\\.)*)"' + "|'((?:[^\\']|\\.)*)'")
    NUMBER = re.compile('[+-]?[0-9]*\.?[0-9]+')
    WHITESPACE = re.compile('\s')

    _parse_order = [STRING, NUMBER, WHITESPACE, COMMA, FALLTHROUGH]

    class Match():
        ENDWHITESPACE = re.compile('\s$')
        UNESCAPE = re.compile('\\\(.)')

        def __init__(self, match, regex):
            self.match = match
            self.regex = regex

        def is_whitespace(self):
            if self.regex == ConfigTokenizer.WHITESPACE:
                return True
            return False

        def is_comma(self):
            if self.regex == ConfigTokenizer.COMMA:
                return True
            return False

        def is_value(self):
            if self.regex == ConfigTokenizer.STRING or \
                    self.regex == ConfigTokenizer.NUMBER or \
                    self.regex == ConfigTokenizer.FALLTHROUGH:
                return True
            return False

        def strip(self, s):
            return self.ENDWHITESPACE.sub('', s)

        def unescape(self, s):
            return self.UNESCAPE.sub(r'\1', s)

        def __str__(self):
            result = ''
            groups = self.match.groups()
            if groups:
                for g in groups:
                    if g is not None:
                        result = g
            else:
                result = self.match.group(0)

            result = self.strip(result)
            if self.regex == ConfigTokenizer.STRING:
                result = self.unescape(result)

            return result

    def __init__(self, s):
        self._consumable = s

    def __iter__(self):
        return self

    def next(self):
        """for python2"""
        return self.__next__()

    def __next__(self):
        for r in self._parse_order:
            m = r.match(self._consumable)
            if m:
                self._consumable = self._consumable[len(m.group(0)):]
                return self.Match(m, r)

        raise StopIteration


class Settings:
    """
    Handles loading/saving settings from/to a file on disk.
    """

    instance = None

    # the different settings in each config block
    general = {
        "windowheight": 550,
        "windowwidth": 900,
        "windowstate": None,
        "bustabs_list": [],
        "addbus_list": [],
        }

    def __init__(self, filename=None):
        """
        Creates a new instance of Settings.

        Parameters:
            filename -- path to the settings file.
                        If None, the default $XDG_CONFIG_HOME/d-feet/config will be used.
        """
        if not filename:
            if 'XDG_CONFIG_HOME' in list(os.environ.keys()):
                self.filename = os.path.join(os.environ['XDG_CONFIG_HOME'], 'd-feet', 'config')
            else:
                self.filename = os.path.join(os.environ['HOME'], '.config', 'd-feet', 'config')
        else:
            self.filename = filename
        self.config = configparser.ConfigParser()

        self.read()

    @classmethod
    def get_instance(cls):
        """This class is a singlton so use this method to get it"""
        if cls.instance:
            return cls.instance

        cls.instance = Settings()
        return cls.instance

    def decode_list(self, s):
        result = []
        lex = ConfigTokenizer(s)
        for item in lex:
            if item.is_value():
                result.append(str(item))

        return result

    def read(self):
        """
        Reads configuration settings from the config file and loads
        then into the Settings dictionaries.
        """
        self.config.read(self.filename)

        if not self.config.has_section("General"):
            self.config.add_section("General")

        for key, value in self.config.items("General"):
            if key.endswith('list'):
                value = self.decode_list(value)

            self.general[key] = value

    def quote(self, value):
        result = ''
        result = re.sub(r'([\\\"\'])', r'\\\1', value)
        return '"%s"' % result

    def write(self):
        """
        Writes configuration settings to the Settings config file.
        """
        for key in self.general:
            if key.endswith('list'):
                for i in range(0, len(self.general[key])):
                    self.general[key][i] = self.quote(self.general[key][i])
                self.general[key] = ','.join(self.general[key])

            if self.general[key] is None:
                self.general[key] = ''

            self.config.set("General", str(key), str(self.general[key]))

        # make sure that the directory that the config file is in exists
        new_file_dir = os.path.split(self.filename)[0]
        if not os.path.isdir(new_file_dir):
            os.makedirs(new_file_dir)
        f = open(self.filename, 'w')
        self.config.write(f)
        f.close()



# -*- coding: utf-8 -*-
import os
from gi.repository import Gtk


class UILoader:
    instance = None

    UI_COUNT = 5

    (UI_MAINWINDOW,
     UI_INTROSPECTION,
     UI_EXECUTEDIALOG,
     UI_ADDCONNECTIONDIALOG,
     UI_BUS
     ) = range(UI_COUNT)

    # {ui_id: ((files,...), root widget)}
    _ui_map = {UI_MAINWINDOW: (('mainwindow.ui',),
                               'appwindow1'),
               UI_INTROSPECTION: (('introspection.ui',),
                                  'box_introspectview'),
               UI_EXECUTEDIALOG: (('executedialog.ui',),
                                  'executedialog1'),
               UI_ADDCONNECTIONDIALOG: (('addconnectiondialog.ui',),
                                        'add_connection_dialog1'),
               UI_BUS: (('bus.ui',),
                        'box_bus')
               }

    def __init__(self, data_dir, ui):
        ui_info = self._ui_map[ui]
        self.ui = Gtk.Builder()
        self.data_dir = data_dir

        # load ui files
        for f in ui_info[0]:
            self.ui.add_from_file(self.ui_dir + '/' + f)

        self.root_widget_name = ui_info[1]

    def get_widget(self, name):
        return self.ui.get_object(name)

    def get_root_widget(self):
        return self.get_widget(self.root_widget_name)

    def connect_signals(self, obj_or_map):
        self.ui.connect_signals(obj_or_map)

    @property
    def ui_dir(self):
        return os.path.join(self.data_dir, "ui")



# -*- coding: utf-8 -*-

from gi.repository import GObject, Gio
from dfeet import dbus_utils


def args_signature_markup(arg_signature):
    return '<small><span foreground="#2E8B57">%s</span></small>' % (arg_signature)


def args_name_markup(arg_name):
    return '<small><span foreground="#000000">%s</span></small>' % (arg_name)


class DBusNode(GObject.GObject):
    """object to represent a DBus Node (object path)"""
    def __init__(self, name, object_path, node_info):
        GObject.GObject.__init__(self)
        self.__name = name
        self.__object_path = object_path
        self.__node_info = node_info  # Gio.GDBusNodeInfo object

    def __repr__(self):
        return "Name: %s ; ObjPath: %s ; NodeInfo: %s" % (
            self.name, self.object_path, self.node_info)

    @property
    def name(self):
        return self.__name

    @property
    def object_path(self):
        return self.__object_path

    @property
    def node_info(self):
        return self.__node_info


class DBusInterface(DBusNode):
    """object to represent a DBus Interface"""
    def __init__(self, dbus_node_obj, iface_info):
        DBusNode.__init__(self, dbus_node_obj.name,
                          dbus_node_obj.object_path, dbus_node_obj.node_info)
        self.__iface_info = iface_info  # Gio.GDBusInterfaceInfo object

    def __repr__(self):
        return "iface '%s' on node '%s'" % (self.iface_info.name, self.node_info.path)

    @property
    def iface_info(self):
        return self.__iface_info


class DBusProperty(DBusInterface):
    """object to represent a DBus Property"""
    def __init__(self, dbus_iface_obj, property_info):
        DBusInterface.__init__(self, dbus_iface_obj, dbus_iface_obj.iface_info)
        self.__property_info = property_info  # Gio.GDBusPropertyInfo object
        self.__value = None  # the value

    def __repr__(self):
        sig = dbus_utils.sig_to_string(self.property_info.signature)
        return "%s %s (%s)" % (sig, self.property_info.name, self.property_info.flags)

    @property
    def property_info(self):
        return self.__property_info

    @property
    def value(self):
        return self.__value

    @value.setter
    def value(self, new_val):
        self.__value = new_val

    @property
    def markup_str(self):
        sig = dbus_utils.sig_to_string(self.property_info.signature)
        readwrite = list()
        if self.readable:
            readwrite.append("read")
        if self.writable:
            readwrite.append("write")
        s = "%s %s <small>(%s)</small>" % (
            args_signature_markup(sig),
            args_name_markup(self.property_info.name), " / ".join(readwrite))
        if self.value is not None:
            s += " = %s" % (self.value,)
        return s

    @property
    def readable(self):
        if int(self.property_info.flags) == int(Gio.DBusPropertyInfoFlags.READABLE) or \
                int(self.property_info.flags) == \
                (int(Gio.DBusPropertyInfoFlags.WRITABLE | Gio.DBusPropertyInfoFlags.READABLE)):
            return True
        else:
            return False

    @property
    def writable(self):
        if int(self.property_info.flags) == int(Gio.DBusPropertyInfoFlags.WRITABLE) or \
                int(self.property_info.flags) == \
                (int(Gio.DBusPropertyInfoFlags.WRITABLE | Gio.DBusPropertyInfoFlags.READABLE)):
            return True
        else:
            return False


class DBusSignal(DBusInterface):
    """object to represent a DBus Signal"""
    def __init__(self, dbus_iface_obj, signal_info):
        DBusInterface.__init__(self, dbus_iface_obj,
                               dbus_iface_obj.iface_info)
        self.__signal_info = signal_info  # Gio.GDBusSignalInfo object

    def __repr__(self):
        return "%s" % (self.signal_info.name)

    @property
    def signal_info(self):
        return self.__signal_info

    @property
    def args(self):
        args = list()
        for arg in self.signal_info.args:
            sig = dbus_utils.sig_to_string(arg.signature)
            args.append({'signature': sig, 'name': arg.name})
        return args

    @property
    def args_markup_str(self):
        result = ''
        result += '<span foreground="#FF00FF">(</span>'
        result += ', '.join('%s' % (args_signature_markup(arg['signature'])) for arg in self.args)
        result += '<span foreground="#FF00FF">)</span>'
        return result

    @property
    def markup_str(self):
        return "%s %s" % (self.signal_info.name, self.args_markup_str)


class DBusMethod(DBusInterface):
    """object to represent a DBus Method"""
    def __init__(self, dbus_iface_obj, method_info):
        DBusInterface.__init__(self, dbus_iface_obj, dbus_iface_obj.iface_info)
        self.__method_info = method_info  # Gio.GDBusMethodInfo object

    def __repr__(self):
        return "%s(%s) ↦ %s (%s)" % (
            self.method_info.name, self.in_args_str,
            self.out_args_str, DBusInterface.__repr__(self))

    @property
    def in_args_code(self):
        in_args = ""
        for a in self.__method_info.in_args:
            in_args += a.signature
        return in_args

    @property
    def method_info(self):
        return self.__method_info

    @property
    def markup_str(self):
        return "%s %s <b>↦</b> %s" % (
            self.method_info.name, self.in_args_markup_str, self.out_args_markup_str)

    @property
    def in_args(self):
        in_args = list()
        for in_arg in self.method_info.in_args:
            sig = dbus_utils.sig_to_string(in_arg.signature)
            in_args.append({'signature': sig, 'name': in_arg.name})
        return in_args

    @property
    def out_args(self):
        out_args = list()
        for out_arg in self.method_info.out_args:
            sig = dbus_utils.sig_to_string(out_arg.signature)
            out_args.append({'signature': sig, 'name': out_arg.name})
        return out_args

    @property
    def in_args_str(self):
        result = ""
        for arg in self.in_args:
            result += "%s %s, " % (arg['signature'], arg['name'])

        return result[0:-2]

    @property
    def out_args_str(self):
        result = ""
        for arg in self.out_args:
            result += "%s %s, " % (arg['signature'], arg['name'])

        return result[0:-2]

    def __args_markup_str(self, args):
        """markup a given list of args"""
        result = ''
        result += '<span foreground="#FF00FF">(</span>'
        result += ', '.join(
            '%s %s' % (
                args_signature_markup(arg['signature']),
                args_name_markup(arg['name'])) for arg in args)
        result += '<span foreground="#FF00FF">)</span>'
        return result

    @property
    def in_args_markup_str(self):
        return self.__args_markup_str(self.in_args)

    @property
    def out_args_markup_str(self):
        return self.__args_markup_str(self.out_args)


class DBusAnnotation(DBusInterface):
    """object to represent a DBus Annotation"""
    def __init__(self, dbus_iface_obj, annotation_info):
        DBusInterface.__init__(self, dbus_iface_obj,
                               dbus_iface_obj.iface_info)
        self.__annotation_info = annotation_info  # Gio.GDBusAnnotationInfo object

    def __repr__(self):
        return "%s: %s" % (self.annotation_info.key, self.annotation_info.value)

    @property
    def annotation_info(self):
        return self.__annotation_info

    @property
    def markup_str(self):
        return "%s: %s" % (self.annotation_info.key, self.annotation_info.value)



# -*- coding: utf-8 -*-

from __future__ import print_function
from gi.repository import Gtk, Gio, GObject, Gdk
from dfeet.window import DFeetWindow
import os


class DFeetApp(Gtk.Application):

    def __init__(self, package, version, data_dir):
        self.package = package
        self.version = version
        self.data_dir = data_dir
        Gtk.Application.__init__(self, application_id="org.gnome.d-feet",
                                 flags=Gio.ApplicationFlags.FLAGS_NONE)

    # Note that the function in C activate() becomes do_activate() in Python
    def do_activate(self):
        self._main_win = DFeetWindow(self, self.package, self.version, self.data_dir)

    # Note that the function in C startup() becomes do_startup() in Python
    def do_startup(self):
        Gtk.Application.do_startup(self)
        builder = Gtk.Builder()
        builder.add_from_file(os.path.join(self.data_dir, "ui", "app-menu.ui"))

        # create actions
        action = Gio.SimpleAction.new("about", None)
        action.connect("activate", self.action_about_cb)
        self.add_action(action)

        action = Gio.SimpleAction.new("help", None)
        action.connect("activate", self.action_help_cb)
        self.add_action(action)

        action = Gio.SimpleAction.new("quit", None)
        action.connect("activate", self.action_quit_cb)
        self.add_action(action)

        self.set_app_menu(builder.get_object("app-menu"))

    def action_quit_cb(self, action, parameter):
        self.quit()

    def action_about_cb(self, action, parameter):
        aboutdialog = DFeetAboutDialog(self.package, self.version)
        aboutdialog.set_transient_for(self._main_win)
        aboutdialog.show()

    def action_help_cb(self, action, parameter):
        screen = Gdk.Screen.get_default()
        link = "help:d-feet"
        Gtk.show_uri(screen, link, Gtk.get_current_event_time())


class DFeetAboutDialog(Gtk.AboutDialog):
    def __init__(self, package, version):
        Gtk.AboutDialog.__init__(self)
        self.set_program_name(package)
        self.set_version(version)
        self.set_license_type(Gtk.License.GPL_2_0)
        self.set_website("https://wiki.gnome.org/Apps/DFeet/")
        self.set_logo_icon_name(package)
        self.connect("response", self.on_close_cb)

    def on_close_cb(self, action, parameter):
        action.destroy()



# -*- coding: utf-8 -*-
# This module facilitates the optional use of libwnck to get application
# icon information. If the wnck module is not installed we fallback to default
# behvior

import gi
from gi.repository import Gtk, Gdk, GLib


def running_in_x11():
    display = Gdk.Display.get_default()
    return display.__gtype__.name == 'GdkX11Display'


try:
    if not running_in_x11():
        raise GLib.Error('Wnck is only meant to be used with X11')
    gi.require_version('Wnck', '3.0')
    from gi.repository import Wnck
    has_libwnck = True
except:
    has_libwnck = False


class IconTable(object):
    instance = None

    def __init__(self):
        # {pid: icon}
        self.app_map = {}

        icon_theme = Gtk.IconTheme.get_default()
        self.default_icon = icon_theme.load_icon('dfeet-icon-default-service', 16, 0)

        if has_libwnck:
            screen = Wnck.Screen.get_default()
            Wnck.Screen.force_update(screen)
            screen.connect('application_opened', self.on_app_open)
            screen.connect('application_closed', self.on_app_close)

            for w in screen.get_windows():
                app = w.get_application()
                pid = app.get_pid()
                icon = app.get_mini_icon()

                if pid not in self.app_map.keys():
                    self.app_map[pid] = icon

    def on_app_open(self, screen, app):
        icon = app.get_mini_icon()
        if icon:
            self.app_map[app.get_pid()] = icon

    def on_app_close(self, screen, app):
        # this is a leak but some apps still exist even if all their
        # top level windows don't.  We need to have a better cleanup
        # based on when an app's services go away
        return
        pid = app.get_pid()

        if pid not in self.app_map.keys():
            del self.app_map[pid]

    def get_icon(self, pid):
        icon = None
        if pid not in self.app_map.keys():
            icon = self.app_map[pid]

        if not icon:
            icon = self.default_icon

        return icon

    @classmethod
    def get_instance(cls):
        if not cls.instance:
            cls.instance = IconTable()

        return cls.instance



# -*- coding: utf-8 -*-
# copyright (C) 2013 Thomas Bechtold <thomasbechtold@jpberlin.de>

# This file is part of D-Feet.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from gi.repository import GObject, Gtk, Gio

import gettext
from gettext import gettext as _
gettext.textdomain('d-feet')

from dfeet.bus_watch import BusWatch
from dfeet.settings import Settings
from dfeet.uiloader import UILoader
from dfeet.addconnectiondialog import AddConnectionDialog
from dfeet.executemethoddialog import ExecuteMethodDialog


class DFeetWindow(Gtk.ApplicationWindow):
    """the main window"""

    HISTORY_MAX_SIZE = 10

    def __init__(self, app, package, version, data_dir):
        Gtk.Window.__init__(self, application=app)
        self.package = package
        self.version = version
        self.data_dir = data_dir
        self.session_bus = None
        self.system_bus = None

        # setup the window
        self.set_default_size(600, 480)
        self.set_icon_name(package)

        # create actions
        action = Gio.SimpleAction.new('connect-system-bus', None)
        action.connect('activate', self.__action_connect_system_bus_cb)
        self.add_action(action)

        action = Gio.SimpleAction.new('connect-session-bus', None)
        action.connect('activate', self.__action_connect_session_bus_cb)
        self.add_action(action)

        action = Gio.SimpleAction.new('connect-other-bus', None)
        action.connect('activate', self.__action_connect_other_bus_cb)
        self.add_action(action)

        # get settings
        settings = Settings.get_instance()
        self.connect('delete-event', self.__delete_cb)
        self.set_default_size(int(settings.general['windowwidth']),
                              int(settings.general['windowheight']))

        # setup ui
        ui = UILoader(self.data_dir, UILoader.UI_MAINWINDOW)
        header = ui.get_widget('headerbar')
        self.set_titlebar(header)
        self.stack = ui.get_widget('buses_stack')
        self.add(self.stack)
        self.__stack_child_added_id = self.stack.connect('add', self.__stack_child_added_cb)
        self.__stack_child_removed_id = self.stack.connect('remove', self.__stack_child_removed_cb)
        self.connect('destroy', self.__on_destroy)

        # create bus history list and load entries from settings
        self.__bus_history = []
        for bus in settings.general['addbus_list']:
            if bus != '':
                self.__bus_history.append(bus)

        # add a System and Session Bus tab
        self.activate_action('connect-system-bus', None)
        self.activate_action('connect-session-bus', None)

        self.show_all()

    @property
    def bus_history(self):
        return self.__bus_history

    @bus_history.setter
    def bus_history(self, history_new):
        self.__bus_history = history_new

    def __stack_child_added_cb(self, stack, child):
        existing = self.lookup_action('close-bus')
        if existing is None:
            action = Gio.SimpleAction.new('close-bus', None)
            action.connect('activate', self.__action_close_bus_cb)
            self.add_action(action)

    def __stack_child_removed_cb(self, stack, child):
        current = self.stack.get_visible_child()
        if current is None:
            self.remove_action('close-bus')

        if child == self.system_bus:
            self.system_bus = None
            # Re-enable the action
            action = Gio.SimpleAction.new('connect-system-bus', None)
            action.connect('activate', self.__action_connect_system_bus_cb)
            self.add_action(action)
        elif child == self.session_bus:
            self.session_bus = None
            # Re-enable the action
            action = Gio.SimpleAction.new('connect-session-bus', None)
            action.connect('activate', self.__action_connect_session_bus_cb)
            self.add_action(action)

    def __on_destroy(self, data=None):
        self.stack.disconnect(self.__stack_child_added_id)
        self.stack.disconnect(self.__stack_child_removed_id)

    def __action_connect_system_bus_cb(self, action, parameter):
        """connect to system bus"""
        try:
            if self.system_bus is not None:
                return
            bw = BusWatch(self.data_dir, Gio.BusType.SYSTEM)
            self.system_bus = bw.box_bus
            self.stack.add_titled(self.system_bus, 'System Bus', 'System Bus')
            self.remove_action('connect-system-bus')
        except Exception as e:
            print(e)

    def __action_connect_session_bus_cb(self, action, parameter):
        """connect to session bus"""
        try:
            if self.session_bus is not None:
                return
            bw = BusWatch(self.data_dir, Gio.BusType.SESSION)
            self.session_bus = bw.box_bus
            self.stack.add_titled(self.session_bus, 'Session Bus', 'Session Bus')
            self.remove_action('connect-session-bus')
        except Exception as e:
            print(e)

    def __action_connect_other_bus_cb(self, action, parameter):
        """connect to other bus"""
        dialog = AddConnectionDialog(self.data_dir, self, self.bus_history)
        result = dialog.run()
        if result == Gtk.ResponseType.OK:
            address = dialog.address
            if address == 'Session Bus':
                self.activate_action('connect-session-bus', None)
                return
            elif address == 'System Bus':
                self.activate_action('connect-system-bus', None)
                return
            else:
                try:
                    bw = BusWatch(self.data_dir, address)
                    self.stack.add_titled(bw.box_bus, address, address)
                    # Fill history
                    if address in self.bus_history:
                        self.bus_history.remove(address)
                    self.bus_history.insert(0, address)
                    # Truncating history
                    if (len(self.bus_history) > self.HISTORY_MAX_SIZE):
                        self.bus_history = self.bus_history[0:self.HISTORY_MAX_SIZE]
                except Exception as e:
                    print("can not connect to '%s': %s" % (address, str(e)))
        dialog.destroy()

    def __action_close_bus_cb(self, action, parameter):
        """close current bus"""
        try:
            current = self.stack.get_visible_child()
            self.stack.remove(current)
        except Exception as e:
            print(e)

    def __delete_cb(self, main_window, event):
        """store some settings"""
        settings = Settings.get_instance()
        size = main_window.get_size()
        pos = main_window.get_position()

        settings.general['windowwidth'] = size[0]
        settings.general['windowheight'] = size[1]

        self.bus_history = self.bus_history[0:self.HISTORY_MAX_SIZE]

        settings.general['addbus_list'] = self.bus_history
        settings.write()
